/*  ----------------------------------------------------------------------------------
 *  triac_controller with dimmer function. on-off or timer application. autostart
 *  or manual start of timer
 *  Triac_dimmer_30, 25.11.2019, by roland stiglmayr, follows v530
 *  signal UP from a button or an optocoupler, decrements phase and so increases power
 *  signal DOWN from a button or an OC, increments phase and so decreases power
 *  
 *  if timer is enabled, timeout after SHUT_OFF_TIME in seconds
 *  
 *  ----------------------------------------------------------------------------------
 *  
 *  
 *  the function of this application is selected by hardware settings according 
 *  to the following table
 *  
 *  ON_OFF_EN        Function
 *  
 *  HIGH             On_Off_App, active with UP_IN = LOW
 *  LOW              Timer_App, timer restarts @ UP_IN = LOW
 *
 *  ON_OFF_EN (on_off enable) on ICSP pin4  
 *  timer timeout after MANUAL_TIME_SEC in seconds
 *  
 *  -----------------------------------------------------------------------------------  
 */


  #define TRIG_OUT_PIN 5                  //arduino style D5
  #define LED_OUT_PIN 7                   //arduino style D7
  #define UP_PIN 9                        //arduino style D9
  #define DOWN_PIN 8                      //arduino style D8
  #define ON_OFF_EN_PIN 11                //arduino style D11, PB3
  #define ZERO_DET_PIN 2                  //arduino style D2= INT0

  // declare I/Os for using the fast macro commands
  //
    #define TRIG_PORT PORTD               //TRIG_OUT @PORTD 
    #define TRIG_OUT  PD5                 //TRIG_OUT @PD5, triac-gate
    #define LED_PORT PORTD                //LED @PORTD
    #define LED_OUT PD7                   //LED @PD7, heartbeat LED
    #define UP_IN_P PINB                  //Signal UP @ PINB
    #define UP_IN PB1                     //low decrements phase angle, starts timer
    #define DOWN_IN_P PINB                //signal DOWN @ PINB
    #define DOWN_IN PB0                   //low increments phase angle
    #define ON_OFF_EN_P PINB              //signal ON_OFF_EN @ PINB
    #define ON_OFF_EN PB3                 //signal selects function
    #define ZERO_DET_P PIND               //ZERO_DET @ PIND
    #define ZERO_DET_IN PD2               //detects mains zero voltage, PD2= INT0

#define EN_INT0 01                        //enables INT0, written to EIMSK
#define INT0_CHANGE 01                    //INTO on logic change -> EICRA
#define INT_CLEAR 03                      //clears pending INT0, INT1 -> EIFRA

#define T1_STOP 00                        //disable counting
#define T1_START 02                       //T1-CK, Tcycle*8= 0.72µs (TCCR1B)
#define T1_TRIGTIME (0xffff-8)            //triggerpulse length 8*.72µs + 4µs = 10µs

#define T2_STOP 00                        //disables counting
#define T2_START 07                       //T2-CK, Tcycle*1024= 92.593µs (TCCR2B)
#define T_ONE_HW 107                      //(107+1) * 92.6µs= 10ms
#define PHASE_MAX  103                    //(103+1) * 92.6µs= 9.63ms
#define PHASE_MIN 1                       //lowest phaseangle, highest output
#define START_PHASE 91                    //start phase after PON, 152° > 150Vpk

#define MAX_CYCLE_COUNT 50                //50 * 20ms= 1 sec
#define SHUT_OFF_TIME (10 * 60)           //shut off time in seconds

boolean Second_HW;                        //set if second halfwave is active
boolean Second_HW_Rdy;                    //period is finished
boolean Shut_Off;                         //shut off triac, load
boolean Idle_Flag= false;                 //start idle mode
boolean Power_Save= false;                //start power save mode
boolean Wd_Rdy= false;
boolean On_Off_App;                       //on_off or timer function

int Cycle_Counter= MAX_CYCLE_COUNT;       //decremented each 20 ms
unsigned int Sec_Counter= 0;              //counts elapsed seconds
unsigned int Shutoff_Time;                //keeps shutoff time @ auto/man
int Phase_Angle= START_PHASE;             //keeps current phase angle

void setup() 
{   
  //  after reset, start with sleep mode to stabilize VCC. prepare the watchdog 
  //  for wakeup
  //
        asm ("WDR");                      //clear wd-timer
        WDTCSR= 0x1c;                     //set WDCE, WDE to enable setting 
        WDTCSR= 0x1c;                     //wd-timer 0.25s      
        WDTCSR= 0x14;                     //clear WDE
        WDTCSR= (WDTCSR | 0x40);          //set WDIE, enable WD-interrupt
        interrupts();                     //enable interrupts      

        SMCR= 0x0f;                       //prepare xtd-standby
        asm ("SLEEP");                    //enter sleep mode
        SMCR= 0x00;                       //finish, not necessary           
        noInterrupts();                   //disable interrupts, after wakeup
        
        MCUSR= (MCUSR & 0xf7);            //clear WDRF, wd-reset active
        WDTCSR= WDTCSR | 0x18;            //set WDCE, enable setting WDE
        WDTCSR= 00;                       //clear WDIE, disables wd-timer
         

  
  digitalWrite (TRIG_OUT_PIN,HIGH);       //avoid glitch
  pinMode (TRIG_OUT_PIN,OUTPUT);          //output to triac-gate
  digitalWrite (TRIG_OUT_PIN,HIGH);       //disable triac
  
  pinMode (ZERO_DET_PIN, INPUT);          //input INT0, zerodetect
  pinMode (LED_OUT_PIN, OUTPUT);          //active state 
  pinMode (UP_PIN, INPUT);                //input signal to decrement phase
  pinMode (DOWN_PIN, INPUT);              //input signal to increment phase
  pinMode (ON_OFF_EN_PIN, INPUT_PULLUP);  //select signal on_off/ timer
    
  ACSR= 0x80;                             //disable analog comparator
                                          //to save power

  TIMSK0= 0;                              //disable timer ints
  TIMSK1= 0;
  TIMSK2= 0;
  
  // init INT0 -----------------------------------------------------
  //  
  EIMSK= 0;                               //disable hw ints
  PCICR= 0;                               //disable pinchange ints
  EICRA= INT0_CHANGE;                     //INT0 on level change
  EIFR= INT_CLEAR;                        //clear pending INT-flags


  //  init 16bit-timer 1, duration of trigger pulse ----------------
  //
  TCCR1A= 0x00;                           //normal operation
  TCCR1B= T1_STOP;                        //stop counter
  TCNT1= T1_TRIGTIME;                     //prepare first cycle
  OCR1AL= 0x00;
  OCR1BL= 0x00;
  TIMSK1= 0x01;                           //enable T1 overflow interrupt
  TIFR1= 0x07;                            //clear pending interrupt flags

  
  //  init 8bit-timer 2, generates second cycle ---------------------
  //
  TCCR2A= 0x02;                           //clear T2 on compareA match
  TCCR2B= T2_STOP;                        //stop T2
  TCNT2= 00;                              //clear counter T2
  OCR2A= T_ONE_HW;                        //one shot one halfwave
  OCR2B= START_PHASE;                     //phase after PON  
  TIMSK2= 0x06;                           //interrupt on A+B match
  TIFR2= 0x07;                            //clear interpt flags


  Second_HW= true;                        //allow synchron start with ZD 
  Second_HW_Rdy= false;                   //starts with INT0   
  Cycle_Counter= MAX_CYCLE_COUNT;         //counts down at 1 second
  Phase_Angle= START_PHASE;               //set start phase
  
  //  test the selected function as set by hardware
  //  ---------------------------------------------
  
  if (bitRead (ON_OFF_EN_P, ON_OFF_EN))
      {
        On_Off_App= true;
        Shut_Off= false;
      }

  Shutoff_Time= SHUT_OFF_TIME;            //prepare shutoff time
      
  pinMode (ON_OFF_EN_PIN, INPUT_PULLUP);  //disable pullup, save power
  Power_Save= true;                       //save power until first cycle
  EIFR= INT_CLEAR;                        //clear pending hwints
 
  interrupts ();                          //now enable interrupts
  EIMSK= EN_INT0;                         //enable the first event the HW interrupt
}

  //-------------------------------------------------------------------
  //  WDT SERVICE ROUTINE, started after each reset, stretches the
  //  reset phase with low power consumption until power is stable
  //  only used once for wake up of sleep mode
  //-------------------------------------------------------------------
  
    ISR (WDT_vect)
        {
          Wd_Rdy=true;
        }


  //---------------------------------------------------------------------
  //  HW INT0 INTERRUPT SERVICE ROUTINE, falling edge at INT0 starts ISR.
  //  following edge is disabled. INT0 driven by AC zero detector, ZD.
  //  immediately trigger if phase = phase_min  
  //---------------------------------------------------------------------

  ISR (INT0_vect)
    { 
     if (Second_HW && !bitRead (ZERO_DET_P, ZERO_DET_IN)) //test if new period starts
                                                          //with low @ zero detect
        {         
          if ((Phase_Angle == PHASE_MIN) && (!Shut_Off))  //if smallest phase
               bitClear (TRIG_PORT, TRIG_OUT);      //start trigger pulse immediately              
          
          TCCR2B= T2_START;                         //start T2, defines second halfwave
          OCR2B= Phase_Angle;                       //set current phase
          bitClear (LED_PORT, LED_OUT);             //LED off                      

          EIMSK= 00;                                //no more hw-trigger while this period
                                                    //protect against transients
          Second_HW= false;                         //first HW is started
          TIMSK2= 06;                               //enable compare match @ a+b
          --Cycle_Counter;                          //decrement each 20 ms
        }
     Power_Save= true;                              //reduce power until T20 
    }

  //-------------------------------------------------------------------
  //  TIMER 2A INTERRUPT SERVICE ROUTINE, called 10 ms after INT0.
  //  starts second halfwave. COMPA-match clears TCNT2 and so 
  //  synchronizes T2B to this event. 
  //  generates trigger pulse if current phase = phase min
  //-------------------------------------------------------------------

    ISR (TIMER2_COMPA_vect)
        {
          if (!Second_HW)                           //test if first HW is active
              {
                if ((Phase_Angle == PHASE_MIN) && (!Shut_Off))
                    bitClear (TRIG_PORT, TRIG_OUT); //start trigger pulse
               
                OCR2B= Phase_Angle;                 //set current phase delay
                Second_HW= true;                    //next event is second HW 
              }       

          Power_Save= true;                        //save power until T2B or INT0
        }


  //---------------------------------------------------------------------
  //  TIMER 2 COMPB ISR, started by INT0 or T2A. starts a trigger pulse
  //  generates further pulses as long as OCR2B < PHASE_MAX by 
  //  preparing further T2B matches. tests on end of second halfwave
  //---------------------------------------------------------------------
  
  ISR (TIMER2_COMPB_vect)
      {
        if (!Shut_Off)                             //test if active
            {
              bitClear (TRIG_PORT, TRIG_OUT);       //start trigger pulse
            }
              
         TCNT1= T1_TRIGTIME;                        //prepare triggertime
         TCCR1B= T1_START;                          //start T1, T1 enables Power_save

         if (OCR2B <= (PHASE_MAX - 4))              //test if current phase < phase max  
              OCR2B= (OCR2B + 4);                   //prepare next fill puls after 370µs
         else
            {                                       //no further 2B match follows
              if (Second_HW)
                {
                  EIFR= INT_CLEAR;                  //clear pending hw-interrupts  
                  EIMSK= EN_INT0;                   //enable INT0 again
                  TCCR2B= T2_STOP;                  //stop T2, no second T2A
                  TCNT2= 0;                         //synchronize T2 to zero-detect
                  Second_HW_Rdy= true;              //period elapsed

                  if ((Cycle_Counter < 5) && (!Shut_Off)) //show active state @ eof second
                      bitSet (LED_PORT, LED_OUT);         //LED on
                }
            }
      }

  //-----------------------------------------------------------------------------
  //  TIMER 1 INTERRUPT SERVICE ROUTINE, finishes trigger pulse
  //-----------------------------------------------------------------------------

    ISR (TIMER1_OVF_vect)
        {
          bitSet (TRIG_PORT, TRIG_OUT);             //finish trigger pulse
          TCCR1B= T1_STOP;                          //stop T1
          Power_Save= true;                         //prepare power-save 
        }


  //------------------------------------------------------------------------------
  //    main loop
  //    tests on requests to change phase, if enabled the timer function is 
  //    realized
  //    power save/ idle mode is activated. sleeps until wake up event 
  //------------------------------------------------------------------------------

  void loop() 
    {
      if (Power_Save)
       {
        if (Shut_Off == true)
            Cycle_Counter= MAX_CYCLE_COUNT;         //prepare next activation
            
        if (Cycle_Counter <= 0)                     //test if one second is finished
          {
            Cycle_Counter= MAX_CYCLE_COUNT;         //preset cycle_counter
             if (!Shut_Off)                      
                 ++Sec_Counter;                     //if not yet ready, inc second            
          }
     

        //  procedure for timer application
        //  -------------------------------

            if (!On_Off_App && Second_HW_Rdy)       //test if timer and end of period
                {
                  Second_HW_Rdy= false;             //finish this period
                  if (Sec_Counter >= Shutoff_Time)  //test if timer is ready
                     {
                      Shut_Off= true;               //disable trigger pulses
                      Sec_Counter= 0;               //prepare restart
                     }
                
                //test on start/ restart condition ---------
                              
                if (!bitRead (PINB, UP_IN) && Shut_Off && (Sec_Counter == 0)) 
                    {
                      Cycle_Counter= MAX_CYCLE_COUNT;   //synchronize cycle_counter
                      Shut_Off= false;              //restart, let current phase
                      Shutoff_Time= SHUT_OFF_TIME;  //prepare shutoff_time
                    }

                if (!bitRead (PINB, UP_IN) && !Shut_Off) //test UP signal
                    {
                      if (Phase_Angle > PHASE_MIN)
                      --Phase_Angle;                //reduce phase of trigger if active
                    }
                else
                    if (!bitRead (PINB, DOWN_IN) && !Shut_Off)  //test DOWN signal
                        {
                          if (Phase_Angle < (PHASE_MAX))
                          ++Phase_Angle;            //increment phase but not shut off
                        }            
              }              
       
        // procedure for ON-OFF application
        // ---------------------------------

           if (On_Off_App && Second_HW_Rdy)         //test if on_off and eof period
              {
                Sec_Counter= 0;                     //clear second counter
                Second_HW_Rdy= false;

                if (!bitRead (PINB, UP_IN))         //test UP signal    
                    {
                      Shut_Off= false;              //enable trigger
                      if (Phase_Angle > PHASE_MIN)
                          --Phase_Angle;            //reduce phase of trigger if active
                    }
                else
                    if (!bitRead (PINB, DOWN_IN))   //test DOWN signal
                        {
                          if (Phase_Angle < PHASE_MAX)
                              {
                               ++Phase_Angle;
                              }
                          else Shut_Off= true;      //at minimum shut off
                     }
              }


      Power_Save= false;        //disables power save after wake up
      SMCR= 0x0f;               //prepare power save
      asm ("SLEEP");
      SMCR= 0x00;               //finish sleep
    }
}

        
            

