/*  -------------------------------------------------------------------------------
 *  mains_controller, switches inductive loads as transformers, limits inrush
 *  current
 *  Triac_softstart_42, 30.11.19, follows V3941, by roland stiglmayr
 *  system-clock = 11.0592 MHz.
 *  uses Watchdog timer interrupt. ports adapted to  hardware. 
 *  ----------------------------------------------------------------------------------
 *
 *  the function of this application is selected by hardware settings according 
 *  to the following table
 *  
 *  ON_OFF_EN   START_D     Function
 *  
 *  HIGH        X           On_Off_App, active while START_IN is LOW
 *  LOW         HIGH        Timer_App, timer starts @ START_IN = LOW
 *                          longer pressing of START_IN doubles time
 *  LOW         LOW         Auto_Timer, timer is started directly at PowerOn                        
 *
 *  START_D (start_direct) on ICSP pin4
 *  timer timeout after MANUAL_TIME_SEC in seconds
 *  auto timeout after AUTO_TIME_SEC in seconds
 *  
 *  --------------------------------------------------------------------------------- */

 
#define TRIG_OUT_PIN 5                    //arduino style D5
#define LED_OUT_PIN 7                     //arduino style D7
#define START_PIN 9                       //arduino style D9
#define ON_OFF_EN_PIN 8                   //arduino style D8, PB0
#define START_D_PIN 11                    //arduino style D11, PB3
#define ZERO_DET_PIN 2                    //arduino style D2= INT0

//declare I/Os for using the fast macro commands
//
  #define TRIG_P PORTD                    //TRIG_OUT @PORTD 
  #define TRIG_OUT  PD5                   //TRIG_OUT @PD5, triac-gate
  #define LED_P PORTD                     //LED @PORTD
  #define LED_OUT PD7                     //LED @PD7, heartbeat LED
  #define START_IN_P PINB                 //Signal START @PINB
  #define START_IN PB1                    //low active signal on-off, start timer
  #define START_D_IN_P PINB               //signal START_D_IN @ PINB
  #define START_D_IN PB3                  //set type of start at timer
  #define ON_OFF_EN_P PINB                //signal ON_OFF_EN @ PINB
  #define ON_OFF_EN PB0                   //signal selects function
  #define ZERO_DET_P PIND                 //ZERO_DET @ PIND
  #define ZERO_DET_IN PD2                 //detects mains zero voltage, PD2= INT0

#define EN_INT0 01                        //enables INT0, written to EIMSK
#define INT0_CHANGE 01                    //INTO on logic change -> EICRA
#define INT_CLEAR 03                      //clears pending INT0, INT1 -> EIFRA

#define T1_STOP 00                        //disable counting
#define T1_START 02                       //T1-CK, Tcycle*8= 0.72µs (TCCR1B)
#define T1_TRIGTIME (0xffff-8)            //triggerpulse length 8*.72µs + 4µs = 10µs
#define T2_STOP 00                        //disables counting
#define T2_START 07                       //T2-CK, Tcycle*1024= 92.593µs (TCCR2B)
#define T_ONE_HW 107                      //(107+1) * 92.6µs= 10ms

#define START_PHASE  90                   //highest phase, (90+1) * 92.6µs= 8.43ms
#define PHASE_MAX  51                     //fill pulses until (51+1) * 92.6µs= 4.8ms (+0.3ms) 
#define PHASE_MIN 0                       //lowest phaseangle
#define PHASE_SKIP 1                      //92.6µs smallest phaseninkrement per period
                                          //per fill pulse
#define MAX_CYCLE_COUNT 50                //50 * 20ms= 1 sec
#define MANUAL_TIME (2*60);               //time in seconds if manual
#define AUTO_TIME (5*60);                 //time in seconds if autotimer

boolean Second_HW;                        //set if second halfwave is active
boolean Second_HW_Rdy;                    //second period to be finished
boolean Shut_Off;                         //shut off triac, load
boolean Power_Save= false;                //start X-standby
boolean Wd_Rdy= false;
boolean Start_Released;                   //input start_in released
boolean On_Off_App=  true;                //keeps type of activated app
boolean Auto_Timer_App= false;            //autostart if timer

int Cycle_Counter= MAX_CYCLE_COUNT;       //decremented each 20 ms
unsigned int Sec_Counter= 0;              //counts ellapsed seconds
unsigned int Shutoff_Time;                //keeps shutoff time
int Phase_Angle= START_PHASE;             //keeps current phase angle
int Time_Compensate;

void setup() 
{   
  noInterrupts ();

  //  after reset, start with sleep mode to stabilize VCC. prepare the watchdog 
  //  for wakeup
  //
        asm ("WDR");                      //clear wd-timer
        WDTCSR= 0x1c;                     //set WDCE, WDE to enable setting 
        WDTCSR= 0x1c;                     //wd-timer 0.25s      
        WDTCSR= 0x14;                     //clear WDE
        WDTCSR= (WDTCSR | 0x40);          //set WDIE, enable WD-interrupt
        interrupts();                     //enable interrupts      

        SMCR= 0x0f;                       //prepare xtd-standby
        asm ("SLEEP");                    //enter sleep mode
        SMCR= 0x00;                       //finish, not necessary           
        noInterrupts();                   //disable interrupts, after wakeup
        
        MCUSR= (MCUSR & 0xf7);            //clear WDRF, wd-reset active
        WDTCSR= WDTCSR | 0x18;            //set WDCE, enable setting WDE
        WDTCSR= 00;                       //clear WDIE, disables wd-timer
  
  digitalWrite (TRIG_OUT_PIN,HIGH);       //avoid glitch
  pinMode (TRIG_OUT_PIN,OUTPUT);          //output to triac-gate
  digitalWrite (TRIG_OUT_PIN,HIGH);       //disable triac
  
  pinMode (ZERO_DET_PIN, INPUT);          //input INT0, zerodetect
  pinMode (LED_OUT_PIN, OUTPUT);          //active state 
  pinMode (START_PIN, INPUT);             //low active, start
  pinMode (ON_OFF_EN_PIN, INPUT);         //select sig on_off/ timer
  pinMode (START_D_PIN, INPUT_PULLUP);    //select sig start directly

  ACSR= 0x80;                             //disable analog comparator
                                          //to save power

  TIMSK0= 0;                              //disable timer ints
  TIMSK1= 0;
  TIMSK2= 0;
  
  // init INT0 -----------------------------------------------------
  //  
  EIMSK= 0;                               //disable hw ints
  PCICR= 0;                               //disable pinchange ints
  EICRA= INT0_CHANGE;                     //INT0 on level change
  EIFR= INT_CLEAR;                        //clear pending INT-flags


  //  init 16bit-timer 1, duration of trigger pulse ----------------
  //
  TCCR1A= 0x00;                           //normal operation
  TCCR1B= T1_STOP;                        //stop counter
  TCNT1= T1_TRIGTIME;                     //prepare first cycle
  OCR1AL= 0x00;
  OCR1BL= 0x00;
  TIMSK1= 0x01;                           //enable T1 overflow interrupt
  TIFR1= 0x07;                            //clear pending interrupt flags

  
  //  init 8bit-timer 2, generates second cycle ---------------------
  //
  TCCR2A= 0x02;                           //clear T2 on compareA match
  TCCR2B= T2_STOP;                        //stop T2
  TCNT2= 00;                              //clear counter T2
  OCR2A= T_ONE_HW;                        //one shot at first halfwave
  OCR2B= START_PHASE;                     //initial phase angle
  TIMSK2= 0x06;                           //interrupt on A+B match
  TIFR2= 0x07;                            //clear interpt flags

  Second_HW= true;                        //allow synchron start with ZD
  Second_HW_Rdy= true;                    //start with INT0  
  Cycle_Counter= MAX_CYCLE_COUNT;         //counts down at 1 second
  Phase_Angle= START_PHASE;               //start with highest phase angle

  //  test the selected function as set by hardware
  //  ---------------------------------------------

  if (bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      bitRead (START_D_IN_P, START_D_IN)) //on_off_en, start_d_in = 11
      On_Off_App= true;                   //ON-OFF selected

  if (bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      !bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 10
      On_Off_App= true;                     //ON-OFF selected

  if (!bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      bitRead (START_D_IN_P, START_D_IN)) //on_off_en, start_d_in = 01
      {                                   //manual timer selected
        On_Off_App= false;
        Auto_Timer_App= false;
      }

  if (!bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      !bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 00
      {                                     //autotimer selected
        On_Off_App= false;
        Auto_Timer_App= true;
      }
      
  if (On_Off_App)                         //test if on-off app is selected
      {
        Shut_Off= true;                   //disable triggerpulses until start
      }
      
  if (!On_Off_App && !Auto_Timer_App)     //manual timer is started
      {
        Shut_Off= true;                   //no pulses until manual start
        Shutoff_Time= MANUAL_TIME;        //prepare shutoff time
        Start_Released= true;             
      }
      
  if (!On_Off_App && Auto_Timer_App)      //auto timer
      {
        Shut_Off= false;                  //start immediately with pulses
        Shutoff_Time= AUTO_TIME;          //prepare shutoff time
        Start_Released= false;            //if not released, no further start
      }          

  pinMode (START_D_PIN, INPUT);           //disable pullup, save power

  Power_Save= true;                       //save power until first cycle
  EIFR= INT_CLEAR;                        //clear pending hwints
  interrupts ();                          //now enable interrupts
  EIMSK= EN_INT0;                         //enable the first event the HW interrupt
}


  //-------------------------------------------------------------------
  //  WDT SERVICE ROUTINE, started after each reset, stretches the
  //  reset phase with low power consumption until power is stable
  //  only used once for wake up of sleep mode
  //-------------------------------------------------------------------
  
    ISR (WDT_vect)
        {
          Wd_Rdy=true;
        }

  //---------------------------------------------------------------------
  //  HW INT0 INTERRUPT SERVICE ROUTINE, falling edge at INT0 starts ISR.
  //  following edge is disabled. INT0 driven by AC zero detector, ZD.  
  //---------------------------------------------------------------------

  ISR (INT0_vect)
    { 
     if (Second_HW && !bitRead (ZERO_DET_P, ZERO_DET_IN)) //test if new period starts
                                                          //with low @ zero detect
        {         
          TCCR2B= T2_START;                         //start T2, defines second halfwave
          OCR2B= Phase_Angle;                       //set current phase angle
          OCR2A= T_ONE_HW;                          //defines eof first HW
          EIMSK= 00;                                //no more hw-trigger while this period
                                                    //protect against transients
          Second_HW= false;                         //first HW is started
          TIMSK2= 06;                               //enable compare match @ a+b
          --Cycle_Counter;                          //decrement each 20 ms
        }
     Power_Save= true;                              //reduce power until T2 or INT0 
    }

  //---------------------------------------------------------------------
  //  TIMER 2A INTERRUPT SERVICE ROUTINE, called 10 ms after INT0. starts
  //  second halfwave time error compensated. COMPA-match clears TCNT2 
  //  and so synchronizes T2B to this event. 
  //  again called at eof protection interval. prepares next period by
  //  setting the phase angle and enabling INT0  
  //---------------------------------------------------------------------

    ISR (TIMER2_COMPA_vect)
        {
          if (!Second_HW)                           //test if first HW is active
              {
                OCR2B= Phase_Angle;                 //set current phase angle
                Second_HW= true;                    //next event is second HW
                OCR2A= (T_ONE_HW - 10);             //reduce 2. HW, the protection interval

                TCCR2B= T2_STOP;                    //longer first HW to compensate time
                TCCR2B= T2_START;                   //now start timer again                                
              }       
          else                                      //finishes this period= 2 HWs
              {
                EIFR= INT_CLEAR;                    //clear pending hw-interrupts  
                EIMSK= EN_INT0;                     //enable INT0 again
                TCCR2B= T2_STOP;                    //stop T2
                TCNT2= 0xff;                        //synchronize T2 to zero-detect
                TIMSK2=06;                          //enable T2COMP a+b interrupt
                Second_HW_Rdy= true;                //period ellapsed

                //if active reduce the current phase angle
                //until PHASE_MIN is reached -------------
                                                    
                if (!Shut_Off && (Phase_Angle > (PHASE_MIN+1))) 
                    Phase_Angle= Phase_Angle - (2 * PHASE_SKIP);  //2 skips
                else
                    {
                    if (!Shut_Off && (Phase_Angle > (PHASE_MIN)))
                        Phase_Angle= PHASE_MIN;                   //smallest                     
                    }
              }
          Power_Save= true;                        //save power until T2B or INT0
        }

  //---------------------------------------------------------------------
  //  TIMER 2 COMPB ISR, started by INT0 or T2A. starts a trigger pulse. 
  //  generates further pulses as long as OCR2B < START_PHASE resp. 
  //  OCR2B < PHASE_MAX by preparing further T2B matches. 
  //---------------------------------------------------------------------
  
  ISR (TIMER2_COMPB_vect)
      {
        if (!Shut_Off)                              //test if active
            {
              bitClear (TRIG_P, TRIG_OUT);          //start trigger pulse
              if (Cycle_Counter < 2)                //show active state @ eof second
                  bitSet (LED_P, LED_OUT);          //LED on
            }
              
         TCNT1= T1_TRIGTIME;                        //prepare triggertime
         TCCR1B= T1_START;                          //start T1, T1 makes power save

         if (OCR2B <= (START_PHASE - PHASE_SKIP))   //test if current phase < start value
            {
              if (OCR2B <= (Phase_Angle + (PHASE_MAX)))   //beetween phaseangle and max           
                  OCR2B= (OCR2B + (4 * PHASE_SKIP));      //prepare next fill pulse 

              if (OCR2B <= (Phase_Angle + (PHASE_MAX/2 +
                            PHASE_MAX/4 + PHASE_MAX/8)))  //beetween phaseangle and n/8 max 
                  {
                    OCR2B= (OCR2B - PHASE_SKIP -          //(4-3)skips
                            PHASE_SKIP - PHASE_SKIP);     //prepare next fill pulse
                  }
            }
      }

  //-----------------------------------------------------------------------------
  //  TIMER 1 INTERRUPT SERVICE ROUTINE, finishes trigger pulse
  //-----------------------------------------------------------------------------

    ISR (TIMER1_OVF_vect)
        {
          bitSet (TRIG_P, TRIG_OUT);                  //finish trigger pulse
          bitClear (LED_P, LED_OUT);                  //LED off
          TCCR1B= T1_STOP;                            //stop T1
          Power_Save= true;                           //prepare X-standby 
        }


//------------------------------------------------------------------------------
//    main loop
//------------------------------------------------------------------------------

void loop() 
{
  if (Power_Save)
     {
      if (Cycle_Counter <= 0)                       //if one second is finished
          {
            Cycle_Counter= MAX_CYCLE_COUNT;         //prepare cycle_counter
             if (!Shut_Off)                      
                 ++Sec_Counter;                     //if not yet ready, inc second
          }

      //  procedure for timer application
      //  -------------------------------
            
      if (!On_Off_App)                              //test on timer_app
        {
          if (Second_HW_Rdy)                        //test eof period
            {
              Second_HW_Rdy= false;                 //prepare first half wave
              
              if (Shut_Off && bitRead (START_IN_P, START_IN)) //test if start is released
                  Start_Released= true;
              
              if (Sec_Counter >= Shutoff_Time)      //test if timer is ready
                {
                  Shut_Off= true;
                  Sec_Counter= 0;                   //prepare restart
                  Phase_Angle= START_PHASE;         //next start with highest phase angle                  
                }
               
              //  test on restart
            
              if (!bitRead (START_IN_P, START_IN) && Shut_Off && (Sec_Counter == 0) && 
                  Start_Released)
                {
                  Cycle_Counter= MAX_CYCLE_COUNT;   //synchronize cycle_counter
                  Shut_Off= false;
                  Start_Released= false;            //no restart until released                              
                  Phase_Angle= START_PHASE;         //start with highest phase angle

                  if (!Auto_Timer_App)              //prepare shutoff_time
                      {Shutoff_Time= MANUAL_TIME;}      
                  else 
                      {Shutoff_Time= AUTO_TIME;}
                }
            }       
        }        
       
        // procedure for ON-OFF application
        // ---------------------------------

          if (On_Off_App)           
            {
              if (Second_HW_Rdy)                    //test eof period
                {
                  Sec_Counter= 0;                   //no second counter
                  Second_HW_Rdy= false;

                    if (!bitRead (START_IN_P, START_IN))   //test enable input
                         Shut_Off= false;
                    else
                       {
                         Shut_Off= true;            //prepare next ON
                         Phase_Angle= START_PHASE;
                       }
                }
            }


      Power_Save= false;                  //continue wo power save
      SMCR= 0x0f;                         //prepare Extended standby
      asm ("SLEEP");
      SMCR= 0x00;                         //finish sleep
    }
}

